@@ -18,6 +18,7 @@ def create_config() -> Config:
     config.deffield('ENABLE_WEB_SERVER', default=False, type_=bool)
     config.deffield('MODE', default=0x0000, desc='CLI or GUI 模式')
     config.deffield('THEME', default='auto', desc='auto/light/dark')
+    config.deffield('ENABLE_NEW_HOMEPAGE', default=False, type_=bool)
     config.deffield('MPV_AUDIO_DEVICE', default='auto', desc='MPV 播放设备')
     config.deffield('COLLECTIONS_DIR',  desc='本地收藏所在目录')
     config.deffield('FORCE_MAC_HOTKEY', desc='强制开启 macOS 全局快捷键功能',
@@ -27,6 +28,13 @@ def create_config() -> Config:
     config.deffield('VIDEO_SELECT_POLICY', default='hd<>')
     config.deffield('ALLOW_LAN_CONNECT', type_=bool, default=False, desc='是否可以从局域网连接服务器')
     config.deffield('PROVIDERS_STANDBY', type_=list, default=None, desc='')
+    config.deffield('ENABLE_YTDL_AS_MEDIA_PROVIDER', type_=bool, default=True,
+                    desc='YTDL 作为备用资源')
+    # For example::
+    #    [{'name': 'match_by_source',
+    #      'source': 'xxx',
+    #      'http_proxy': 'http://127.0.0.1:7890'},]
+    config.deffield('YTDL_RULES', type_=list, default=None, desc='')
     # TODO(cosven): maybe
     # 1. when it is set to 2, find standby from other providers first.
     # 2. when it is set to 3, play it's MV model instead of using MV's media.
