@@ -5,29 +5,33 @@ import random
 from enum import IntEnum, Enum
 from typing import Optional, TYPE_CHECKING
 
-from feeluown.excs import ModelNotFound, ProviderIOError
+from feeluown.excs import ProviderIOError
 from feeluown.utils import aio
 from feeluown.utils.aio import run_fn, run_afn
 from feeluown.utils.dispatch import Signal
 from feeluown.utils.utils import DedupList
-from feeluown.player import Metadata, MetadataFields
 from feeluown.library import (
-    MediaNotFound, SongModel, ModelType, ResourceNotFound, VideoModel,
+    MediaNotFound, SongModel, ModelType, VideoModel, ModelNotFound,
+    BriefSongModel,
 )
 from feeluown.media import Media
-from feeluown.library import reverse
+from .metadata_assembler import MetadataAssembler
 
 if TYPE_CHECKING:
     from feeluown.app import App
 
 logger = logging.getLogger(__name__)
 
+TASK_SET_CURRENT_MODEL = 'playlist.set_current_model'
+TASK_PLAY_MODEL = 'playlist.play_model'
+TASK_PREPARE_MEDIA = 'playlist.prepare_media'
+
 
 class PlaybackMode(IntEnum):
     """
     Playlist playback mode.
 
-    .. versiondeprecated:: 3.8.12
+    .. deprecated:: 3.8.12
         Please use PlaylistRepeatMode and PlaylistShuffleMode instead.
     """
     one_loop = 0  #: One Loop
@@ -73,6 +77,14 @@ class PlaylistMode(IntEnum):
     fm = 1  #: FM mode
 
 
+class PlaylistPlayModelStage(IntEnum):
+    prepare_media = 10
+    find_standby_by_mv = 20
+    find_standby = 30
+    prepare_metadata = 40
+    load_media = 50
+
+
 class Playlist:
     def __init__(self, app: 'App', songs=None, playback_mode=PlaybackMode.loop,
                  audio_select_policy='hq<>'):
@@ -81,6 +93,7 @@ class Playlist:
         :param playback_mode: :class:`feeluown.player.PlaybackMode`
         """
         self._app = app
+        self._metadata_mgr = MetadataAssembler(app)
 
         #: init playlist mode normal
         self._mode = PlaylistMode.normal
@@ -125,8 +138,6 @@ class Playlist:
         #: When watch mode is on, playlist try to play the mv/video of the song
         self.watch_mode = False
 
-        self._t_scm = self._app.task_mgr.get_or_create('set-current-model')
-
         # .. versionadded:: 3.7.11
         #    The *songs_removed* and *songs_added* signal.
         self.songs_removed = Signal()  # (index, count)
@@ -134,6 +145,9 @@ class Playlist:
         # .. versionadded:: 3.9.0
         #    The *play_model_handling* signal.
         self.play_model_handling = Signal()
+        # .. versionadded:: 4.1.7
+        #    The *play_model_handling* signal.
+        self.play_model_stage_changed = Signal()
 
         self._app.player.media_finished.connect(self._on_media_finished)
         self.song_changed.connect(self._on_song_changed)
@@ -299,7 +313,7 @@ class Playlist:
     def clear(self):
         """remove all songs from playlists"""
         if self.current_song is not None:
-            self.current_song = None
+            self.set_current_song_none()
         length = len(self._songs)
         self._songs.clear()
         if length > 0:
@@ -327,7 +341,7 @@ class Playlist:
         """从播放列表中获取一首可以播放的歌曲
 
         :param base: base index
-        :param random: random strategy or not
+        :param random_: random strategy or not
         :param direction: forward if > 0 else backward
         :param loop: regard the song list as a loop
 
@@ -381,13 +395,16 @@ class Playlist:
             next_song = self._get_good_song(random_=True)
         else:
             current_index = self._songs.index(self.current_song)
-            if current_index == len(self._songs) - 1:
-                if self.playback_mode in (PlaybackMode.loop, PlaybackMode.one_loop):
-                    next_song = self._get_good_song()
-                elif self.playback_mode == PlaybackMode.sequential:
-                    next_song = None
+            is_last_song = current_index == len(self._songs) - 1
+            if is_last_song and self.playback_mode == PlaybackMode.sequential:
+                next_song = None
             else:
-                next_song = self._get_good_song(base=current_index+1, loop=False)
+                if is_last_song:
+                    base_index = 0
+                else:
+                    base_index = current_index + 1
+                loop = self.playback_mode != PlaybackMode.sequential
+                next_song = self._get_good_song(base=base_index, loop=loop)
         return next_song
 
     @property
@@ -459,38 +476,27 @@ class Playlist:
     def current_song_mv(self) -> Optional[VideoModel]:
         return self._current_song_mv
 
-    def set_current_song(self, song) -> Optional[asyncio.Task]:
-        """设置当前歌曲，将歌曲加入到播放列表，并发出 song_changed 信号
-
-        .. note::
-
-            该方法理论上只应该被 Player 对象调用。
-
-        if song has not valid media, we find a replacement in other providers
-
-        .. versionadded:: 3.7.11
-           The method is added to replace current_song.setter.
-        """
-        if song is None:
-            self.pure_set_current_song(None, None, None)
-            return None
-
-        if self.mode is PlaylistMode.fm and song not in self._songs:
-            self.mode = PlaylistMode.normal
-
-        # FIXME(cosven): `current_song.setter` depends on app.task_mgr and app.library,
-        # which make it hard to test.
-        return self._t_scm.bind_coro(self.a_set_current_song(song))
-
     async def a_set_current_song(self, song):
         """Set the `song` as the current song.
 
         If the song is bad, then this will try to use a standby in Playlist.normal mode.
         """
+        if song is None:
+            self.set_current_song_none()
+            return None
+
+        if self.mode is PlaylistMode.fm and song not in self._songs:
+            self.mode = PlaylistMode.normal
+
         target_song = song  # The song to be set.
         media = None        # The corresponding media to be set.
         try:
-            media = await self._prepare_media(song)
+            self.play_model_stage_changed.emit(PlaylistPlayModelStage.prepare_media)
+            media = await self._app.task_mgr.run_afn_preemptive(
+                self._prepare_media,
+                song,
+                name=TASK_PREPARE_MEDIA,
+            )
         except MediaNotFound as e:
             if e.reason is MediaNotFound.Reason.check_children:
                 await self.a_set_current_song_children(song)
@@ -510,12 +516,14 @@ class Playlist:
         # The song has no media, try to find and use standby unless it is in fm mode.
         if media is None:
             if self._app.config.ENABLE_MV_AS_STANDBY:
-                self._app.show_msg('尝试获取音乐视频的播放资源...')
+                self.play_model_stage_changed.emit(
+                    PlaylistPlayModelStage.find_standby_by_mv)
                 media = await self._prepare_mv_media(song)
 
             if media:
                 self._app.show_msg('使用音乐视频作为其播放资源 ✅')
             else:
+                self._app.show_msg('未找到可用的音乐视频资源 🙁')
                 # if mode is fm mode, do not find standby song, just skip the song.
                 if self.mode is PlaylistMode.fm:
                     self.mark_as_bad(song)
@@ -524,10 +532,15 @@ class Playlist:
 
                 logger.info(f"no media found for {song}, mark it as bad")
                 self.mark_as_bad(song)
+                self.play_model_stage_changed.emit(PlaylistPlayModelStage.find_standby)
                 target_song, media = await self.find_and_use_standby(song)
 
-        metadata = await self._prepare_metadata_for_song(target_song)
-        self.pure_set_current_song(target_song, media, metadata)
+        metadata = None
+        if media is not None:
+            self.play_model_stage_changed.emit(PlaylistPlayModelStage.prepare_metadata)
+            metadata = await self._metadata_mgr.prepare_for_song(target_song)
+        self.play_model_stage_changed.emit(PlaylistPlayModelStage.load_media)
+        self.set_current_song_with_media(target_song, media, metadata)
 
     async def a_set_current_song_children(self, song):
         # TODO: maybe we can just add children to playlist?
@@ -564,108 +577,42 @@ class Playlist:
         self._app.show_msg(f'未找到 {song} 的备用歌曲')
         return song, None
 
-    def pure_set_current_song(self, song, media, metadata=None):
+    def set_current_song_with_media(self, song, media, metadata=None):
         if song is None:
-            self._current_song = None
-        else:
-            # add it to playlist if song not in playlist
-            if song in self._songs:
-                self._current_song = song
-            else:
-                self.insert(song)
-                self._current_song = song
+            self.set_current_song_none()
+            return
+        # Add it to playlist if song not in playlist.
+        if song not in self._songs:
+            self.insert(song)
+        self._current_song = song
         self.song_changed.emit(song)
         self.song_changed_v2.emit(song, media)
-
-        if song is not None:
-            if media is None:
-                run_afn(self.a_next)
-            else:
-                # Note that the value of model v1 {}_display may be None.
-                kwargs = {}
-                if not self._app.has_gui:
-                    kwargs['video'] = False
-                # TODO: set artwork field
-                self._app.player.play(media, metadata=metadata, **kwargs)
+        if media is None:
+            self._app.show_msg("没找到可用的播放链接，播放下一首...")
+            run_afn(self.a_next)
         else:
-            self._app.player.stop()
+            kwargs = {}
+            if not self._app.has_gui:
+                kwargs['video'] = False
+            # TODO: set artwork field
+            self._app.player.play(media, metadata=metadata, **kwargs)
 
-    async def _prepare_metadata_for_song(self, song):
-        metadata = Metadata({
-            MetadataFields.uri: reverse(song),
-            MetadataFields.source: song.source,
-            MetadataFields.title: song.title_display or '',
-            # The song.artists_name should return a list of strings
-            MetadataFields.artists: [song.artists_name_display or ''],
-            MetadataFields.album: song.album_name_display or '',
-        })
-        try:
-            song: SongModel = await aio.wait_for(
-                aio.run_fn(self._app.library.song_upgrade, song),
-                timeout=1,
-            )
-        except ResourceNotFound:
-            return metadata
-        except:  # noqa
-            logger.exception(f"fetching song's meta failed, song:'{song.title_display}'")
-            return metadata
-
-        artwork = song.pic_url
-        released = song.date
-        if not (artwork and released) and song.album is not None:
-            try:
-                album = await aio.wait_for(
-                    aio.run_fn(self._app.library.album_upgrade, song.album),
-                    timeout=1
-                )
-            except ResourceNotFound:
-                pass
-            except:  # noqa
-                logger.warning(
-                    f"fetching song's album meta failed, song:{song.title_display}")
-            else:
-                artwork = album.cover or artwork
-                released = album.released or released
-                # For model v1, the cover can be a Media object.
-                # For example, in fuo_local plugin, the album.cover is a Media
-                # object with url set to fuo://local/songs/{identifier}/data/cover.
-                if isinstance(artwork, Media):
-                    artwork = artwork.url
-
-        # Try to use album meta first.
-        if artwork and released:
-            metadata[MetadataFields.artwork] = artwork
-            metadata[MetadataFields.released] = released
-        else:
-            metadata[MetadataFields.artwork] = song.pic_url or artwork
-            metadata[MetadataFields.released] = song.date or released
-        return metadata
-
-    async def _prepare_metadata_for_video(self, video):
-        metadata = Metadata({
-            # The value of model v1 title_display may be None.
-            MetadataFields.title: video.title_display or '',
-            MetadataFields.source: video.source,
-            MetadataFields.uri: reverse(video),
-        })
-        try:
-            video = await aio.run_fn(self._app.library.video_upgrade, video)
-        except ModelNotFound as e:
-            logger.warning(f"can't get cover of video due to {str(e)}")
-        else:
-            metadata[MetadataFields.artwork] = video.cover
-        return metadata
+    def set_current_song_none(self):
+        """A special case of `set_current_song_with_media`."""
+        self._current_song = None
+        self.song_changed.emit(None)
+        self.song_changed_v2.emit(None, None)
+        self._app.player.stop()
 
     async def _prepare_media(self, song):
-        task_spec = self._app.task_mgr.get_or_create('prepare-media')
-        # task_spec.disable_default_cb()
         if self.watch_mode is True:
-            mv_media = await task_spec.bind_coro(self._prepare_mv_media(song))
+            mv_media = await self._prepare_mv_media(song)
             if mv_media:
                 return mv_media
             self._app.show_msg('未找到可用的歌曲视频资源')
-        return await task_spec.bind_blocking_io(
-            self._app.library.song_prepare_media, song, self.audio_select_policy)
+        return await aio.run_fn(
+            self._app.library.song_prepare_media, song, self.audio_select_policy,
+        )
 
     async def _prepare_mv_media(self, song) -> Optional[Media]:
         try:
@@ -680,25 +627,17 @@ class Playlist:
             logger.exception(f'fail to get {song} mv: {e}')
         return mv_media
 
-    def set_current_model(self, model):
-        """
-        .. versionadded: 3.7.13
-        """
-        if model is None:
-            self._app.player.stop()
-            return
-        if ModelType(model.meta.model_type) is ModelType.song:
-            return self.set_current_song(model)
-        return self._t_scm.bind_coro(self.a_set_current_model(model))
-
     async def a_set_current_model(self, model):
         """
         TODO: handle when model is a song
 
         .. versionadded: 3.7.13
         """
-        assert ModelType(model.meta.model_type) is ModelType.video, \
-            "{model.meta.model_type} is not supported, expecting a video model, "
+        if model is None:
+            self._app.player.stop()
+            return
+        if isinstance(model, BriefSongModel):
+            return await self.a_set_current_song(model)
 
         video = model
         try:
@@ -710,34 +649,73 @@ class Playlist:
         except MediaNotFound:
             self._app.show_msg('没有可用的播放链接')
         else:
-            metadata = await self._prepare_metadata_for_video(video)
+            metadata = await self._metadata_mgr.prepare_for_video(video)
             kwargs = {}
             if not self._app.has_gui:
                 kwargs['video'] = False
             self._app.player.play(media, metadata=metadata, **kwargs)
 
+    async def a_play_model(self, model):
+        """
+        .. versionadded: 4.1.7
+        """
+        # Stop the player so that user know the action is working.
+        self._app.player.stop()
+        if model is None:
+            return
+        self.play_model_handling.emit()
+        if ModelType(model.meta.model_type) is ModelType.song:
+            fn = self.a_set_current_song
+            upgrade_fn = self._app.library.song_upgrade
+        else:
+            fn = self.a_set_current_model
+            upgrade_fn = self._app.library.video_upgrade
+        try:
+            # Try to upgrade the model.
+            model = await aio.run_fn(upgrade_fn, model)
+        except ModelNotFound:
+            pass
+        except:  # noqa
+            logger.exception(f'upgrade model:{model} failed')
+        try:
+            await self._app.task_mgr.run_afn_preemptive(
+                fn, model, name=TASK_SET_CURRENT_MODEL
+            )
+        except:  # noqa
+            logger.exception('play model failed')
+        else:
+            self._app.player.resume()
+            logger.info(f'play a model ({model}) succeed')
+
     """
     Sync methods.
 
     Currently, playlist has both async and sync methods to keep backward
     compatibility. Sync methods will be replaced by async methods in the end.
+    Sync methods just wrap the async method.
     """
     def play_model(self, model):
         """Set current model and play it
 
         .. versionadded: 3.7.14
         """
-        # Stop the player so that user know the action is working.
-        self._app.player.stop()
-        self.play_model_handling.emit()
-        task = self.set_current_model(model)
-        if task is not None:
-            def cb(future):
-                try:
-                    future.result()
-                except:  # noqa
-                    logger.exception('play model failed')
-                else:
-                    self._app.player.resume()
-                    logger.info(f'play a model ({model}) succeed')
-            task.add_done_callback(cb)
+        self._app.task_mgr.run_afn_preemptive(
+            self.a_play_model, model, name=TASK_PLAY_MODEL
+        )
+
+    def set_current_model(self, model) -> asyncio.Task:
+        """
+        .. versionadded: 3.7.13
+        """
+        return self._app.task_mgr.run_afn_preemptive(
+            self.a_set_current_model, model, name=TASK_SET_CURRENT_MODEL,
+        )
+
+    def set_current_song(self, song):
+        """
+        .. versionadded:: 3.7.11
+           The method is added to replace current_song.setter.
+        """
+        return self._app.task_mgr.run_afn_preemptive(
+            self.a_set_current_song, song, name=TASK_SET_CURRENT_MODEL
+        )
